import base64
import re
from types import AsyncGeneratorType
from typing import Union, Any, List
from w3lib.url import url_query_cleaner

import bs4
from yarl import URL

from feedsearch_crawler.crawler import Crawler, Item, Request, Response

from feedsearch_crawler.feed_spider.dupefilter import NoQueryDupeFilter
from feedsearch_crawler.feed_spider.feed_info import FeedInfo
from feedsearch_crawler.feed_spider.feed_info_parser import FeedInfoParser
from feedsearch_crawler.feed_spider.site_meta import SiteMeta
from feedsearch_crawler.feed_spider.site_meta_parser import SiteMetaParser


class FeedsearchSpider(Crawler):
    duplicate_filter_class = NoQueryDupeFilter
    htmlparser = "html.parser"
    favicon_data_uri = True
    try_urls: Union[List[str], bool] = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.site_meta_processor = SiteMetaParser(self)
        self.feed_info_parser = FeedInfoParser(self)
        self.site_metas = set()
        self.favicons = dict()
        self.post_crawl_callback = self.populate_feed_site_meta
        if "try_urls" in kwargs:
            self.try_urls = kwargs["try_urls"]
        if "favicon_data_uri" in kwargs:
            self.favicon_data_uri = kwargs["favicon_data_uri"]

    async def parse(self, request: Request, response: Response) -> AsyncGeneratorType:
        """
        Parse a Response for feeds or site metadata.

        :param request: Request
        :param response: Response
        :return: AsyncGenerator yielding Items, Requests, or iterative AsyncGenerators
        """

        # If the Response is not OK then there's no data to parse.
        if not response.ok:
            return

        # If the Response contains JSON then attempt to parse it as a JsonFeed.
        if response.json:
            if "version" and "jsonfeed" and "feed_url" in response.json:
                yield self.feed_info_parser.parse_item(request, response, type="json")
                return

        if not isinstance(response.text, str):
            self.logger.debug("No text in %s", response)
            return

        url_origin = response.url.origin()
        request_url_origin = request.url.origin()
        # If the returned url is an origin url, or the request url is an origin url (and there may have been a redirect)
        # then parse the site meta.
        if response.url == url_origin or request.url == request_url_origin:
            yield self.site_meta_processor.parse_item(request, response)

        # Restrict the RSS check to the first 500 characters, otherwise it's almost definitely not an actual feed.
        data = response.text.lower()[:500]

        if bool(data.count("<rss") + data.count("<rdf") + data.count("<feed")):
            yield self.feed_info_parser.parse_item(request, response, type="xml")
            return

        # Make sure the Response XML has been parsed if it exists.
        soup = await response.xml
        if not soup:
            return

        # Find all links in the Response.
        links = soup.find_all(tag_has_href)
        for link in links:
            # Follow all valid links if they are a valid "alternate" link (RSS Feed Discovery) or
            # if they look like they might point to valid feeds.
            if should_follow_url(link, response):
                yield self.follow(link.get("href"), self.parse, response)

    async def parse_xml(self, response_text: str) -> Any:
        """
        Parse Response text as XML.
        Used to allow implementations to provide their own XML parser.

        :param response_text: Response text as string.
        :return: None
        """
        return bs4.BeautifulSoup(response_text, self.htmlparser)

    async def process_item(self, item: Item) -> None:
        """
        Process parsed FeedInfo or SiteMeta items.

        :param item: Item object
        :return: None
        """
        if isinstance(item, FeedInfo):
            self.items.add(item)
        elif isinstance(item, SiteMeta):
            self.site_metas.add(item)

    # noinspection PyPep8
    async def populate_feed_site_meta(self) -> None:
        """
        Populate FeedInfo site information with data from the relevant SiteMeta item
        """
        for feed in self.items:
            # Check each SiteMeta for a url host match
            for meta in self.site_metas:
                # If the meta url host begins with www, then we remove it because the feed may be on
                # a different subdomain
                host = meta.url.host
                if host.startswith("www."):
                    host = host[len("www.") :]

                # If the meta url host is in the feed url host then we can assume that the feed belongs to that site
                if host in feed.url.host:
                    feed.site_url = meta.url
                    feed.site_name = meta.site_name
                    if not feed.favicon:
                        feed.favicon = meta.icon_url

            # Populate favicon data uri if available
            if feed.favicon:
                feed.favicon_data_uri = self.favicons.get(feed.favicon, "")

    async def create_data_uri(self, request: Request, response: Response) -> None:
        """
        Create a data uri from a favicon image.

        :param request: Request
        :param response: Response
        :return: None
        """
        if not response.ok or not isinstance(response.data, bytes):
            return

        try:
            encoded = base64.b64encode(response.data)
            uri = "data:image/png;base64," + encoded.decode(response.encoding)
            self.favicons[request.url] = uri
        except Exception as e:
            self.logger.warning("Failure encoding image: %s", e)

    def create_start_urls(self, url: Union[str, URL]) -> List[URL]:
        """
        Create URLs for the initial Requests.

        :param url: Original query URL
        :return: List of URLs to search
        """
        if isinstance(url, str):
            if "//" not in url:
                url = f"//{url}"
            url = URL(url)

        if url.scheme not in ["http", "https"]:
            url = url.with_scheme("http")

        origin = url.origin()

        urls = [url, origin]

        # Common paths for feeds.
        suffixes = {
            "index.xml",
            "atom.xml",
            "feeds",
            "feeds/default",
            "feed",
            "feed/default",
            "feeds/posts/default",
            "?feed=rss",
            "?feed=atom",
            "?feed=rss2",
            "?feed=rdf",
            "rss",
            "atom",
            "rdf",
            "index.rss",
            "index.rdf",
            "index.atom",
            "data/rss",
            "rss.xml",
            "index.json",
            "about",
            "about/feeds",
            "rss-feeds",
        }

        if self.try_urls:
            if isinstance(self.try_urls, list):
                urls.extend(origin.join(URL(suffix)) for suffix in self.try_urls)
            else:
                urls.extend(origin.join(URL(suffix)) for suffix in suffixes)

        return urls


def should_follow_url(link: bs4.Tag, response: Response) -> bool:
    """
    Check that the link should be followed if it may contain feed information.

    :param link: Link tag.
    :param response: Response
    :return: boolean
    """
    href: str = link.get("href")
    link_type: str = link.get("type")

    # No href value in link.
    if not href:
        return False

    url = URL(href)

    is_one_jump: bool = is_one_jump_from_original_domain(url, response)

    # If the link may have a valid feed type then follow it regardless of the url text.
    if (
        link_type
        and any(
            map(link_type.lower().count, ["application/json", "rss", "atom", "rdf"])
        )
        and "json+oembed" not in link_type
        and is_one_jump
    ):
        return True
    # Else validate the actual URL string for possible feed values.
    elif (
        is_one_jump
        and not has_invalid_contents(href)
        and not is_invalid_filetype(href)
        and is_feedlike_url(url, href)
        and not has_comments_in_querystring(url)
    ):
        return True

    return False


def tag_has_href(tag: bs4.Tag) -> bool:
    """
    Find all tags that contain links.

    :param tag: XML tag
    :return: boolean
    """
    return tag.has_attr("href")


def is_one_jump_from_original_domain(url: Union[str, URL], response: Response) -> bool:
    """
    Check that the current URL is only one response away from the originally queried domain.

    We want to be able to follow potential feed links that point to a different domain than
    the originally queried domain, but not to follow any deeper than that.

    Sub-domains of the original domain are ok.

    i.e: the following are ok
        "test.com" -> "feedhost.com"
        "test.com/feeds" -> "example.com/feed.xml"
        "test.com" -> "feeds.test.com"

    not ok:
        "test.com" -> "feedhost.com" (we stop here) -> "feedhost.com/feeds"

    :param url: URL object or string
    :param response: Response object
    :return: boolean
    """
    if isinstance(url, str):
        url = URL(url)

    if not url.is_absolute():
        url = url.join(response.url)

    # This is the first Response in the chain
    if len(response.history) == 1:
        return True

    # URL is same domain
    if url.host == response.history[0].host:
        return True

    # URL is sub-domain
    if response.history[0].host in url.host:
        return True

    # URL domain and current Response domain are different from original domain
    if (
        response.history[-1].host != response.history[0].host
        and url.host != response.history[0].host
    ):
        return False

    return True


# Regex to check URL string for invalid file types.
file_regex = re.compile(".(jpe?g|png|gif|bmp|mp4|mp3|mkv|md|css|avi)/?$", re.IGNORECASE)


def is_invalid_filetype(url: str) -> bool:
    """
    Check if url string has an invalid filetype extension.

    :param url: URL string
    :return: boolean
    """
    if file_regex.search(url.strip()):
        return True
    return False


def has_comments_in_querystring(url: URL) -> bool:
    """
    Check if URL querystring contains comment keys.

    :param url: URL object
    :return: boolean
    """
    return any(key in url.query for key in ["comment", "comments", "post"])


# Regex to check that a feed-like string is a whole word to help rule out false positives.
feedlike_regex = re.compile("\\b(rss|feed|atom|json|xml|rdf|feeds)\\b", re.IGNORECASE)


def is_feedlike_url(url: URL, url_string: str) -> bool:
    """
    Check if url looks like it may point to something resembling a feed.

    :param url: URL object
    :param url_string: URL string
    :return: boolean
    """
    # Check url string without query parameters
    if feedlike_regex.search(url_query_cleaner(url_string)):
        return True
    # Check querystring keys
    for key in url.query:
        if feedlike_regex.search(key):
            return True
    return False


def has_invalid_contents(string: str) -> bool:
    """
    Ignore any string containing the following strings.

    :param string: String to check
    :return: boolean
    """
    return any(
        map(
            string.lower().count,
            ["wp-includes", "wp-content", "wp-json", "xmlrpc", "wp-admin", "/amp/"],
        )
    )



import asyncio
import copy
import inspect
import logging
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from statistics import harmonic_mean
from types import AsyncGeneratorType
from typing import List, Any
from typing import Union

import aiohttp
from aiohttp import ClientTimeout
from yarl import URL

from feedsearch_crawler.crawler.duplicatefilter import DuplicateFilter
from feedsearch_crawler.crawler.item import Item
from feedsearch_crawler.crawler.lib import coerce_url, ignore_aiohttp_ssl_eror
from feedsearch_crawler.crawler.request import Request
from feedsearch_crawler.crawler.response import Response

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    uvloop = None
    pass


@dataclass
class CallbackResult:
    """Dataclass for holding callback results and recording recursion"""

    result: Any
    callback_recursion: int


class Crawler(ABC):

    # Class Name of the Duplicate Filter.
    # May be overridden to use different Duplicate Filter.
    # Not an instantiation of the class.
    duplicate_filter_class = DuplicateFilter

    # Callback to be run after all workers are finished.
    post_crawl_callback = None

    # Max number of concurrent http requests.
    concurrency: int = 10
    # Max size of incoming http response content.
    max_content_length = 1024 * 1024 * 10
    # Max crawl depth. i.e. The max length of the response history.
    max_depth: int = 0
    # Max callback recursion depth, to prevent accidental infinite recursion from AsyncGenerators.
    max_callback_recursion: int = 10

    # List of worker tasks.
    _workers = []

    # ClientSession for requests. Created on Crawl start.
    _session: aiohttp.ClientSession
    # Task queue for Requests. Created on Crawl start.
    _request_queue: asyncio.Queue
    # Semaphore for controlling HTTP Request concurrency.
    _semaphore: asyncio.Semaphore

    def __init__(
        self,
        start_urls: List[str] = None,
        concurrency: int = 10,
        total_timeout: Union[float, ClientTimeout] = 10,
        request_timeout: Union[float, ClientTimeout] = 3,
        user_agent: str = "",
        max_content_length: int = 1024 * 1024 * 10,
        max_depth: int = 10,
        headers: dict = None,
        allowed_schemes: List[str] = None,
        *args,
        **kwargs,
    ):
        """
        Base class for a WebCrawler implementation.

        :param allowed_schemes: List of strings of allowed Request URI schemes. e.g. ["http", "https"]
        :param start_urls: List of initial URLs to crawl.
        :param concurrency: Max number of workers and of concurrent HTTP requests.
        :param total_timeout: Total aiohttp ClientSession timeout. Crawl will end if this timeout is triggered.
        :param request_timeout: Total timeout for each individual HTTP request.
        :param user_agent: Default User-Agent for HTTP requests.
        :param max_content_length: Max size in bytes of incoming http response content.
        :param max_depth: Max crawl depth. i.e. The max length of the response history.
        :param headers: Default HTTP headers to be included in each request.
        :param args: Additional positional arguments for subclasses.
        :param kwargs: Additional keyword arguments for subclasses.
        """
        self.start_urls = start_urls or []
        self.concurrency = concurrency

        if not isinstance(total_timeout, ClientTimeout):
            total_timeout = aiohttp.ClientTimeout(total=total_timeout)
        if not isinstance(request_timeout, ClientTimeout):
            request_timeout = aiohttp.ClientTimeout(total=request_timeout)

        self.total_timeout: ClientTimeout = total_timeout
        self.request_timeout: ClientTimeout = request_timeout

        self.max_content_length = max_content_length
        self.max_depth = max_depth

        self.user_agent = user_agent or (
            "Mozilla/5.0 (compatible; Feedsearch-Crawler; +https://pypi.org/project/feedsearch-crawler)"
        )

        self.headers = {"User-Agent": self.user_agent, "Upgrade-Insecure-Requests": "1"}

        if headers:
            self.headers = {**self.headers, **headers}

        self.allowed_schemes = allowed_schemes

        self.logger = logging.getLogger("feedsearch_crawler")

        # Default set for parsed items.
        self.items: set = set()

        # URL Duplicate Filter instance.
        self._duplicate_filter = self.duplicate_filter_class()

        self.request_durations = []
        self.request_content_length = []

        # Crawl statistics.
        self.stats: dict = {
            # Number of HTTP Requests added and sent.
            "requests_added": 0,
            # Number of HTTP Requests that were successful (HTTP Status code 200-299).
            "requests_successful": 0,
            # Number of HTTP Requests that were unsuccessful (HTTP Status code not in 200s).
            "requests_failed": 0,
            # Total size in bytes of all HTTP Requests.
            "content_length_total": 0,
            # Harmonic mean of total HTTP Request content length in bytes.
            "content_length_avg": 0,
            # Highest HTTP Request content length in bytes.
            "content_length_max": 0,
            # Lowest HTTP Request content length in bytes.
            "content_length_min": 0,
            # Number of Items processed.
            "items_processed": 0,
            # Number of URls seen and added to duplicate filter.
            "urls_seen": 0,
            # Harmonic mean of HTTP Request duration in Milliseconds.
            "requests_duration_avg": 0,
            # Highest HTTP request duration in Milliseconds.
            "requests_duration_max": 0,
            # Lowest HTTp request duration in Milliseconds.
            "requests_duration_min": 0,
            # Total HTTP request duration in Milliseconds.
            "requests_duration_total": 0,
            # Total duration of crawl in Milliseconds.
            "total_duration": 0,
        }

    async def _handle_request(self, request: Request) -> None:
        """
        Handle fetching of Requests and processing of Request callbacks.

        :param request: Request
        :return: None
        """
        try:
            if request.has_run:
                self.logger.warning("%s has already run", request)
                return

            start = time.perf_counter()

            # Fetch the request and run its callback
            results, response = await request.fetch_callback(self._semaphore)

            dur = int((time.perf_counter() - start) * 1000)
            self.request_durations.append(dur)
            self.logger.debug(
                "Fetched: url=%s dur=%dms status=%s prev=%s",
                response.url,
                dur,
                response.status_code,
                response.originator_url,
            )

            if response.ok:
                self.stats["requests_successful"] += 1
            else:
                self.stats["requests_failed"] += 1

            self.request_content_length.append(response.content_length)

            # Mark the Response URL as seen in the duplicate filter, as it may be different from the Request URL
            # due to redirects.
            await self._duplicate_filter.url_seen(response.url, response.method)

            # Add callback results to the queue for processing.
            if results:
                self._request_queue.put_nowait(CallbackResult(results, 0))

        except asyncio.CancelledError:
            self.logger.debug("Cancelled: %s", request)
        except Exception as e:
            self.logger.exception("Exception during %s: %s", request, e)
        finally:
            return

    async def _process_request_callback_result(
        self, result: Any, callback_recursion: int = 0
    ) -> None:
        """
        Process the Request callback result depending on the result type.
        Request callbacks may contain nested iterators.

        :param result: Callback Result. May be an CallbackResult class, AsyncGenerator, Coroutine, Request, or Item.
        :param callback_recursion: Incremented counter to limit this method's recursion.
        :return: None
        """
        if callback_recursion >= self.max_callback_recursion:
            self.logger.warning(
                "Max callback recursion of %d reached", self.max_callback_recursion
            )
            return

        try:
            # If a CallbackResult class is passed, process the result values from within the class.
            if isinstance(result, CallbackResult):
                await self._process_request_callback_result(
                    result.result, result.callback_recursion
                )
            # For async generators, put each value back on the queue for processing.
            # This will happen recursively until the end of the recursion chain or max_callback_recursion is reached.
            elif inspect.isasyncgen(result):
                async for value in result:
                    self._request_queue.put_nowait(
                        CallbackResult(value, callback_recursion + 1)
                    )
            # For coroutines, await the result then put the value back on the queue for further processing.
            elif inspect.iscoroutine(result):
                value = await result
                self._request_queue.put_nowait(
                    CallbackResult(value, callback_recursion + 1)
                )
            # Requests are checked for uniqueness and put onto the queue.
            elif isinstance(result, Request):
                await self._process_request(result)

            # Items are handled by the implementing Class
            elif isinstance(result, Item):
                await self.process_item(result)
                self.stats["items_processed"] += 1
        except Exception as e:
            self.logger.exception(e)

    async def _process_request(self, request: Request) -> None:
        """
        Process a Request onto the Request Queue.

        Before a Request is added to the Queue, first check that the Request URL has not already been seen,
        that the max URL depth has not been reached, and that the URI scheme is allowed.

        :param request: HTTP Request
        :return: None
        """
        # If URL is not already seen, and add it to the duplicate filter seen list.
        if await self._duplicate_filter.url_seen(request.url, request.method):
            return

        # The URL scheme must be in the list of allowed schemes.
        if self.allowed_schemes and request.url.scheme not in self.allowed_schemes:
            self.logger.debug(
                "URI Scheme '%s' not allowed: %s", request.url.scheme, request
            )
            return

        # Restrict the depth of the Request chain to the maximum depth.
        if self.max_depth and len(request.history) >= self.max_depth:
            self.logger.debug("Max Depth of '%d' reached: %s", self.max_depth, request)
            return

        self.stats["requests_added"] += 1
        self.logger.debug("Queue Add: %s", request)
        # Add the Request to the queue for processing.
        self._request_queue.put_nowait(request)

    def follow(
        self, url: Union[str, URL], callback=None, response: Response = None, **kwargs
    ) -> Request:
        """
        Follow a URL by creating an HTTP Request.

        If the URL is not absolute then it is joined with the previous Response URL.
        The previous Response history is copied to the Request.

        :param url: URL to follow.
        :param callback: Callback method to run if the Request is successful.
        :param response: Previous Response that contained the Request URL.
        :param kwargs: Optional Request keyword arguments. See Request for details.
        :return: Request
        """
        if isinstance(url, str):
            url = URL(url)

        history = []
        if response:
            # Join the URL to the Response URL if it doesn't contain a domain.
            if not url.is_absolute():
                url = response.url.origin().join(url)
            # Copy the Response history so that it isn't a pointer.
            history = copy.deepcopy(response.history)

        request = Request(
            url=url,
            request_session=self._session,
            history=history,
            callback=callback,
            xml_parser=self.parse_xml,
            max_content_length=self.max_content_length,
            timeout=self.request_timeout,
            **kwargs,
        )

        return request

    @abstractmethod
    async def process_item(self, item: Item) -> None:
        """
        Processed a parsed Item in some way. e.g. Add it to the Item set, or database, or send a signal.

        :param item: A parsed Item.
        """
        self.items.add(item)

    @abstractmethod
    async def parse_xml(self, response_text: str) -> Any:
        """
        Parse Response text as XML.
        Used to allow implementations to provide their own XML parser.

        :param response_text: Response text as string.
        """
        raise NotImplementedError("Not Implemented")

    @abstractmethod
    async def parse(self, request: Request, response: Response) -> AsyncGeneratorType:
        """
        Parse an HTTP Response. Must yield Items, Requests, AsyncGenerators, or Coroutines.

        :param request: HTTP Request that created the Response.
        :param response: HTTP Response.
        """
        raise NotImplementedError("Not Implemented")

    async def _work(self):
        """
        Worker function for handling request queue items.
        """
        while True:
            item = await self._request_queue.get()

            try:
                # Fetch Request and handle callbacks
                if isinstance(item, Request):
                    try:
                        await self._handle_request(item)
                    except asyncio.CancelledError:
                        self.logger.debug("Cancelled Request: %s", item)
                # Process Callback results
                elif isinstance(item, CallbackResult):
                    await self._process_request_callback_result(
                        item.result, item.callback_recursion
                    )
            finally:
                self._request_queue.task_done()

    async def _run_callback(self, callback, *args, **kwargs) -> None:
        """
        Runs a callback function.

        :param callback: Function to run. May be async.
        :param args: Positional arguments to pass to the function.
        :param kwargs: Keyword arguments to pass to the function.
        :return: None
        """
        if not callback:
            return
        if inspect.iscoroutinefunction(callback):
            await callback(*args, **kwargs)
        elif inspect.isfunction(callback):
            callback(*args, **kwargs)
        else:
            self.logger.warning("Callback %s must be a coroutine or function", callback)

    def create_start_urls(self, url: Union[str, URL]) -> List[URL]:
        """
        Create the start URLs for the crawl from an initial URL. May be overridden.

        :param url: Initial URL
        """
        if isinstance(url, str):
            url = URL(url)

        if url.scheme not in ["http", "https"]:
            url = url.with_scheme("http")

        return [url]

    async def crawl(self, url: Union[URL, str] = ""):
        """
        Start the web crawler.

        :param url: An optional URL to start the crawl. If not provided then start_urls are used.
        """

        # Fix for ssl errors
        ignore_aiohttp_ssl_eror(asyncio.get_running_loop())

        if url:
            self.start_urls = self.create_start_urls(url)

        if not self.start_urls:
            raise ValueError("crawler.start_urls are required")

        start = time.perf_counter()
        # Create the Request Queue and ClientSession within the asyncio loop.
        self._request_queue = asyncio.Queue()

        # Create the Semaphore for controlling HTTP Request concurrency within the asyncio loop.
        self._semaphore = asyncio.Semaphore(self.concurrency)

        self._session = aiohttp.ClientSession(
            timeout=self.total_timeout, headers=self.headers
        )

        # Create a Request for each start URL and add it to the Request Queue.
        for url in self.start_urls:
            await self._process_request(self.follow(coerce_url(url), self.parse))

        # Create workers to process the Request Queue.
        # Create twice as many workers as potential concurrent requests, to handle request callbacks without delay.
        self._workers = [
            asyncio.create_task(self._work()) for _ in range(self.concurrency * 2)
        ]

        try:
            # Run workers within the ClientSession.
            async with self._session:
                await asyncio.wait_for(
                    self._request_queue.join(), timeout=self.total_timeout.total
                )
        except asyncio.TimeoutError:
            self.logger.debug("Timed out after %s seconds", self.total_timeout)
        finally:
            # Make sure all workers are cancelled.
            for w in self._workers:
                w.cancel()

        # Run the post crawl callback if it exists.
        await self._run_callback(self.post_crawl_callback)

        # The ClientSession is closed only after all work is completed.
        await self._session.close()

        duration = int((time.perf_counter() - start) * 1000)

        # Record statistics
        self.stats["total_duration"] = duration
        self.stats["requests_duration_total"] = int(sum(self.request_durations))
        self.stats["requests_duration_avg"] = int(harmonic_mean(self.request_durations))
        self.stats["requests_duration_max"] = int(max(self.request_durations))
        self.stats["requests_duration_min"] = int(min(self.request_durations))
        self.stats["content_length_total"] = int(sum(self.request_content_length))
        self.stats["content_length_avg"] = int(
            harmonic_mean(self.request_content_length)
        )
        self.stats["content_length_max"] = int(max(self.request_content_length))
        self.stats["content_length_min"] = int(min(self.request_content_length))
        self.stats["urls_seen"] = len(self._duplicate_filter.fingerprints)

        self.logger.info(
            "Crawl finished: requests=%s time=%dms",
            (self.stats["requests_failed"] + self.stats["requests_successful"]),
            duration,
        )
        self.logger.debug("Stats: %s", self.stats)



from typing import Any, Union, Dict
from yarl import URL


def coerce_url(url: Union[URL, str], https: bool = False) -> URL:
    """
    Coerce URL to valid format

    :param url: URL
    :param https: Force https if no scheme in url
    :return: str
    """
    if isinstance(url, str):
        url = URL(url)

    scheme = "https" if https else "http"

    if not url.is_absolute():
        url_string = str(url)
        split = url_string.split("/", 1)
        url = URL.build(scheme=scheme, host=split[0])
        if len(split) > 1:
            url = url.with_path(split[1])

    return url


def to_bytes(text, encoding: str = "utf-8", errors: str = "strict"):
    """Return the binary representation of `text`. If `text`
    is already a bytes object, return it as-is."""
    if not text:
        return b""
    if isinstance(text, bytes):
        return text
    return text.encode(encoding, errors)


def to_string(item: Any, encoding: str = "utf-8", errors: str = "strict") -> str:
    """
    Return the string representation of 'item'.
    """
    if item is None:
        return ""
    if isinstance(item, bytes):
        return item.decode(encoding, errors)
    return str(item)


def case_insensitive_key(key: str, dictionary: Dict) -> bool:
    """
    Check if a case-insensitive key is in a dictionary.
    """
    k = key.lower()
    for key in dictionary.keys():
        if key.lower() == k:
            return True


def ignore_aiohttp_ssl_eror(loop, aiohttpversion="3.5.4"):
    """Ignore aiohttp #3535 issue with SSL data after close
     There appears to be an issue on Python 3.7 and aiohttp SSL that throws a
    ssl.SSLError fatal error (ssl.SSLError: [SSL: KRB5_S_INIT] application data
    after close notify (_ssl.c:2609)) after we are already done with the
    connection. See GitHub issue aio-libs/aiohttp#3535
     Given a loop, this sets up a exception handler that ignores this specific
    exception, but passes everything else on to the previous exception handler
    this one replaces.
     If the current aiohttp version is not exactly equal to aiohttpversion
    nothing is done, assuming that the next version will have this bug fixed.
    This can be disabled by setting this parameter to None
     """
    import ssl
    import aiohttp
    import asyncio

    try:
        import uvloop

        protocol_class = uvloop.loop.SSLProtocol
    except ImportError:
        protocol_class = asyncio.sslproto.SSLProtocol
        pass

    if aiohttpversion is not None and aiohttp.__version__ != aiohttpversion:
        return

    orig_handler = loop.get_exception_handler()

    # noinspection PyUnresolvedReferences
    def ignore_ssl_error(loop, context):
        errors = ["SSL error", "Fatal error"]
        if any(x in context.get("message") for x in errors):
            # validate we have the right exception, transport and protocol
            exception = context.get("exception")
            protocol = context.get("protocol")
            if (
                isinstance(exception, ssl.SSLError)
                and exception.reason == "KRB5_S_INIT"
                and isinstance(protocol, protocol_class)
            ):
                if loop.get_debug():
                    asyncio.log.logger.debug("Ignoring aiohttp SSL KRB5_S_INIT error")
                return
        if orig_handler is not None:
            orig_handler(loop, context)
        else:
            loop.default_exception_handler(context)

    loop.set_exception_handler(ignore_ssl_error)



import time
from datetime import datetime, date
from types import AsyncGeneratorType
from typing import Tuple, List, Union, Dict

import feedparser
import udatetime
from bs4 import BeautifulSoup
from yarl import URL

from feedsearch_crawler.crawler import ItemParser, Request, Response, to_string
from feedsearch_crawler.feed_spider.feed_info import FeedInfo
from feedsearch_crawler.feed_spider.lib import parse_header_links


class FeedInfoParser(ItemParser):
    async def parse_item(
        self, request: Request, response: Response, *args, **kwargs
    ) -> AsyncGeneratorType:
        self.logger.info("Parsing feed %s", response.url)

        content_type = response.headers.get("content-type", "")

        item = FeedInfo(response.url, content_type)

        # Check link headers first for WebSub content discovery
        # https://www.w3.org/TR/websub/#discovery
        if response.headers:
            item.hubs, item.self_url = self.header_links(response.headers)

        if "type" not in kwargs:
            raise ValueError("type keyword argument is required")

        try:
            data_type = kwargs["type"]
            if data_type == "json":
                item.content_type = "application/json"
                self.parse_json(item, response.json)
            elif data_type == "xml":
                self.parse_xml(item, response.data, response.encoding, response.headers)
                if not item.content_type:
                    item.content_type = "text/xml"
        except Exception as e:
            self.logger.exception("Failed to parse feed %s, Error: %s", item, e)

        if item.favicon and self.spider.favicon_data_uri:
            yield self.spider.follow(item.favicon, self.spider.create_data_uri)

        item.content_length = response.content_length
        self.score_item(item, response.history[0])
        yield item

    def parse_xml(
        self, item: FeedInfo, data: str, encoding: str, headers: Dict
    ) -> None:
        """
        Get info from XML (RSS or ATOM) feed.
        """

        # Parse data with feedparser
        # Don't wrap this in try/except, feedparser eats errors and returns bozo instead
        parsed = self.parse_raw_data(data, encoding, headers)
        if not parsed or parsed.get("bozo") == 1:
            if not isinstance(
                parsed.get("bozo_exception"), feedparser.CharacterEncodingOverride
            ):
                item.bozo = 1
                self.logger.warning("No valid feed data for %s", item)
                return

        feed = parsed.get("feed")
        if not feed:
            return

        # Only search if no hubs already present from headers
        if not item.hubs:
            item.hubs, item.self_url = self.websub_links(feed)

        if item.hubs and item.self_url:
            item.is_push = True

        item.version = parsed.get("version")
        item.title = self.feed_title(feed)
        item.description = self.feed_description(feed)

        try:
            dates = []
            now_date = datetime.utcnow().date()

            for entry in parsed.get("entries", None):
                if entry.get("published_parsed"):
                    published = datetime.fromtimestamp(
                        time.mktime(entry.get("published_parsed", ""))
                    )
                    if published.date() <= now_date:
                        dates.append(published)

                if entry.get("updated_parsed"):
                    updated = datetime.fromtimestamp(
                        time.mktime(entry.get("updated_parsed", ""))
                    )
                    if updated.date() <= now_date:
                        dates.append(updated)

            item.last_updated = sorted(dates, reverse=True)[0]
        except Exception as e:
            self.logger.error("Unable to get feed published date: %s", e)
            pass

    def parse_json(self, item: FeedInfo, data: dict) -> None:
        """
        Get info from JSON feed.

        :param item: FeedInfo object
        :param data: JSON object
        :return: None
        """
        item.version = data.get("version")
        if "https://jsonfeed.org/version/" not in item.version:
            item.bozo = 1
            return

        item.title = data.get("title")
        item.description = data.get("description")

        favicon = data.get("favicon")
        if favicon:
            item.favicon = URL(favicon)

        # Only search if no hubs already present from headers
        if not item.hubs:
            try:
                item.hubs = list(hub.get("url") for hub in data.get("hubs", []))
            except (IndexError, AttributeError):
                pass

        if item.hubs:
            item.is_push = True

        try:
            dates = []
            now_date: date = datetime.utcnow().date()

            for entry in data.get("items"):
                if entry.get("date_published"):
                    published: datetime = udatetime.from_string(entry["date_published"])
                    if published.date() <= now_date:
                        dates.append(published)

                if entry.get("date_modified"):
                    modified: datetime = udatetime.from_string(entry["date_modified"])
                    if modified.date() <= now_date:
                        dates.append(modified)

            item.last_updated = sorted(dates, reverse=True)[0]
        except Exception as e:
            self.logger.error("Unable to get feed published date: %s", e)
            pass

    def parse_raw_data(
        self, raw_data: Union[str, bytes], encoding: str = "utf-8", headers: Dict = None
    ) -> Dict:
        """
        Loads the raw RSS/Atom XML data.
        Returns feedparser Dict.
        https://pythonhosted.org/feedparser/

        :param raw_data: RSS/Atom XML feed
        :type raw_data: str
        :param encoding: Character encoding of raw_data
        :type encoding: str
        :param headers: Response headers
        :return: Dict
        """
        if not encoding:
            encoding = "utf-8"

        h = {}
        if headers:
            if isinstance(headers, dict):
                h = headers
            else:
                try:
                    h.update({k.lower(): v for (k, v) in headers.items()})
                except KeyError:
                    pass

            h.pop("content-encoding", None)

        try:
            start = time.perf_counter()

            if isinstance(raw_data, str):
                raw_data: bytes = raw_data.encode(encoding)

            content_length = len(raw_data)

            # We want to pass data into feedparser as bytes, otherwise if we accidentally pass a url string
            # it will attempt a fetch
            data = feedparser.parse(raw_data, response_headers=h)

            dur = int((time.perf_counter() - start) * 1000)
            self.logger.debug("Feed Parse: size=%s dur=%sms", content_length, dur)

            return data
        except Exception as e:
            self.logger.exception("Could not parse RSS data: %s", e)

    def feed_title(self, feed: dict) -> str:
        """
        Get feed title

        :param feed: feed dict
        :return: str
        """
        title = feed.get("title", None)
        if not title:
            return ""
        return self.clean_title(title)

    def clean_title(self, title: str) -> str:
        """
        Cleans title string, and shortens if too long.
        Have had issues with dodgy feed titles.

        :param title: Title string
        :return: str
        """
        try:
            title = BeautifulSoup(title, self.spider.htmlparser).get_text()
            if len(title) > 1024:
                title = title[:1020] + "..."
            return title
        except Exception as ex:
            self.logger.exception("Failed to clean title: %s", ex)
            return ""

    @staticmethod
    def feed_description(feed: dict) -> str:
        """
        Get feed description.

        :param feed: feed dict
        :return: str
        """
        subtitle = feed.get("subtitle", None)
        if subtitle:
            return subtitle
        return feed.get("description", None)

    @staticmethod
    def websub_links(feed: dict) -> Tuple[List[str], str]:
        """
        Returns a tuple containing the hub url and the self url for
        a parsed feed.

        :param feed: An RSS feed parsed by feedparser
        :type feed: dict
        :return: tuple
        """
        links = feed.get("links", [])
        return FeedInfoParser.find_hubs_and_self_links(links)

    @staticmethod
    def header_links(headers: dict) -> Tuple[List[str], str]:
        """
        Attempt to get self and hub links from HTTP headers
        https://www.w3.org/TR/websub/#x4-discovery

        :param headers: Dict of HTTP headers
        :return: None
        """
        link_header = headers.get("Link")
        links: list = []
        if link_header:

            links = parse_header_links(to_string(link_header))
        return FeedInfoParser.find_hubs_and_self_links(links)

    @staticmethod
    def find_hubs_and_self_links(links: List[dict]) -> Tuple[List[str], str]:
        """
        Parses a list of links into self and hubs urls

        :param links: List of parsed HTTP Link Dicts
        :return: Tuple
        """
        hub_urls: List[str] = []
        self_url: str = ""

        if not links:
            return [], ""

        for link in links:
            try:
                if link["rel"] == "hub":
                    href: str = link["href"]
                    hub_urls.append(href)
                elif link["rel"] == "self":
                    self_url = link["href"]
            except KeyError:
                continue

        return hub_urls, self_url

    @staticmethod
    def score_item(item: FeedInfo, original_url: URL):
        score = 0

        url_str = str(item.url)

        # -- Score Decrement --

        if original_url:
            host = original_url.host
            if host.startswith("www."):
                host = host[len("www.") :]

            if host not in item.url.host:
                score -= 20

        # Decrement the score by every extra path in the url
        parts_len = len(item.url.parts)
        if parts_len > 2:
            score -= (parts_len - 2) * 2

        if item.bozo:
            score -= 20
        if not item.description:
            score -= 10
        if "georss" in url_str:
            score -= 10
        if "alt" in url_str:
            score -= 7
        if "comments" in url_str or "comments" in item.title.lower():
            score -= 15

        # -- Score Increment --
        if item.url.scheme == "https":
            score += 10
        if item.is_push:
            score += 10
        if "index" in url_str:
            score += 15

        kw = ["atom", "rss", ".xml", "feed", "rdf"]
        for p, t in zip(range(len(kw) * 2, 0, -2), kw):
            if t in url_str:
                score += p

        item.score = score
